@@ -9,6 +9,7 @@ import pyuv
 import datetime
 import errno
 import logging
+import os
 import time
 
 try:
@@ -41,7 +42,6 @@ class IOLoop(object):
         if impl is not None:
             raise RuntimeError('When using pyuv the poller implementation cannot be specifiedi')
         self._loop = _loop or pyuv.Loop()
-        self._poll_handles = {}
         self._handlers = {}
         self._callbacks = deque()
         self._callback_lock = thread.allocate_lock()
@@ -79,31 +79,30 @@ class IOLoop(object):
 
     def _close_loop_handles(self):
         def cb(handle):
-            try:
+            if not handle.closed:
                 handle.close()
-            except Exception:
-                pass
         self._loop.walk(cb)
 
-    def close(self, all_fds=False, all_handlers=False):
-        if all_fds:
-            for fd in self._handlers:
-                poll, stack = self._handlers[fd]
-                if not poll.closed:
-                    poll.close()
-            self._handlers = {}
+    def close(self):
+        for fd, (poll, callback) in self._handlers.items():
+            poll.close()
 
-        if all_handlers:
-            self._close_loop_handles()
-            # Run the loop so the close callbacks are fired and memory is freed
-            # It will not block because all handles are closed
-            assert not self._loop.run_once(), "there are pending handles"
+            try:
+                os.close(fd)
+            except Exception:
+                logging.debug("error closing fd %s", fd, exc_info=True)
+
+
+        self._handlers = {}
+        # Run the loop so the close callbacks are fired and memory is freed
+        self._loop.run(pyuv.UV_RUN_NOWAIT)
+        self._loop = None
 
     def add_handler(self, fd, handler, events):
         if fd in self._handlers:
             raise IOError("fd %d already registered" % fd)
+
         poll = pyuv.Poll(self._loop, fd)
-        poll.fd = fd
         self._handlers[fd] = (poll, stack_context.wrap(handler))
         poll_events = 0
         if (events & IOLoop.READ):
@@ -122,7 +121,9 @@ class IOLoop(object):
         poll.start(poll_events, self._handle_poll_events)
 
     def remove_handler(self, fd):
-        self._handlers.pop(fd, None)
+        items = self._handlers.pop(fd, None)
+        if items is not None:
+            items[0].close()
 
     def set_blocking_signal_threshold(self, seconds, action):
         raise NotImplementedError
@@ -133,22 +134,28 @@ class IOLoop(object):
     def log_stack(self, signal, frame):
         raise NotImplementedError
 
-    def start(self, run_loop=True):
+    def start(self):
         if self._stopped:
             self._stopped = False
             return
         self._thread_ident = thread.get_ident()
         self._running = True
-        if run_loop:
-            while self._running:
-                # We should use run() here, but we need to have break() for that
-                self._loop.run_once()
-            # reset the stopped flag so another start/stop pair can be issued
-            self._stopped = False
+
+        # This timer will prevent busy-looping in case there is nothing scheduled in the loop
+        timer = pyuv.Timer(self._loop)
+        timer.start(lambda x: None, 3600, 3600)
+
+        self._loop.run(pyuv.UV_RUN_DEFAULT)
+
+        timer.close()
+
+        # reset the stopped flag so another start/stop pair can be issued
+        self._stopped = False
 
     def stop(self):
         self._running = False
         self._stopped = True
+        self._loop.stop()
         self._waker.wake()
 
     def running(self):
@@ -196,13 +203,16 @@ class IOLoop(object):
     def _handle_poll_events(self, handle, poll_events, error):
         events = 0
         if error is not None:
-            # TODO: do I need to do anything else here?
-            events |= IOLoop.ERROR
-        if (poll_events & pyuv.UV_READABLE):
+            # Some error was detected, signal readability and writability so that the
+            # handler gets and handles the error
             events |= IOLoop.READ
-        if (poll_events & pyuv.UV_WRITABLE):
             events |= IOLoop.WRITE
-        fd = handle.fd
+        else:
+            if (poll_events & pyuv.UV_READABLE):
+                events |= IOLoop.READ
+            if (poll_events & pyuv.UV_WRITABLE):
+                events |= IOLoop.WRITE
+        fd = handle.fileno()
         try:
             self._handlers[fd][1](fd, events)
         except (OSError, IOError) as e:
@@ -230,16 +240,17 @@ class _Timeout(object):
     __slots__ = ['deadline', 'callback', 'io_loop', '_timer']
 
     def __init__(self, deadline, callback, io_loop=None):
+        now = time.time()
         if (isinstance(deadline, six.integer_types)
                 or isinstance(deadline, float)):
             self.deadline = deadline
         elif isinstance(deadline, datetime.timedelta):
-            self.deadline = time.time() + _Timeout.timedelta_to_seconds(deadline)
+            self.deadline = now + _Timeout.timedelta_to_seconds(deadline)
         else:
             raise TypeError("Unsupported deadline %r" % deadline)
         self.callback = callback
         self.io_loop = io_loop or IOLoop.instance()
-        timeout = max(self.deadline - time.time(), 0)
+        timeout = max(self.deadline - now, 0)
         self._timer = pyuv.Timer(self.io_loop._loop)
         self._timer.start(self._timer_cb, timeout, 0.0)
 
